#include <mex.h>
#include "RemoteCaptury.cpp"

#include <unordered_map>

static bool connected = false;

static void connect(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	int ret = 0;

	if (connected) {
		mexPrintf("already connected\n");
		goto out;
	}

	if (nrhs < 2) {
		mexPrintf("not enough arguments\n");
		goto out;
	}

	if (!mxIsChar(prhs[1])) {
		mexPrintf("the second argument needs to be a character array.\n");
		goto out;
	}

	if (mxGetM(prhs[1]) != 1) {
		mexPrintf("the second argument must be a 1D array.\n");
		goto out;
	}

	{
	std::vector<char> address(mxGetN(prhs[1]) + 1, '\0');
	strncpy(address.data(), mxArrayToString(prhs[1]), address.size()-1);

	int port;
	if (nrhs >= 3) {
		if (!mxIsScalar(prhs[2])) {
			mexPrintf("the third argument must be a scalar.\n");
			goto out;
		}

		port = (int)mxGetScalar(prhs[2]);
	} else
		port = 2101;
	if (Captury_connect(address.data(), port) == 0) {
		mexPrintf("failed to connect.\n");
		goto out;
	}

	if (Captury_synchronizeTime() == 0) {
		mexPrintf("failed to synchronize time.\n");
		goto out;
	}

	ret = 1;
	connected = true;
	}
out:
	if (nlhs >= 1)
		plhs[0] = mxCreateLogicalScalar(ret);
}

static std::unordered_map<int, std::vector<float>> recordedDataPerActor;
static std::unordered_map<int, std::vector<uint64_t>> recordedTimestampsPerActor;
static std::unordered_map<int, int> numTransformsPerActor;

static void newPose(CapturyActor* actor, CapturyPose* pose, int trackingQuality)
{
	std::vector<uint64_t>& timestamps = recordedTimestampsPerActor[pose->actor];
	if (timestamps.empty() || timestamps.back() != pose->timestamp) {
		std::vector<float>& data = recordedDataPerActor[pose->actor];
		data.resize(data.size() + pose->numTransforms * 6);
		memcpy(data.data() + data.size() - pose->numTransforms * 6, pose->transforms, pose->numTransforms * 6 * sizeof(float));
		timestamps.push_back(pose->timestamp);
		numTransformsPerActor[pose->actor] = pose->numTransforms;
	}
}

static int64_t recordingStartTime = 0;

static void startRecording(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	int ret = 0;
	// if (Captury_synchronizeTime() == 0) {
	// 	mexPrintf("failed to synchronize time\n");
	// 	goto out;
	// }

	recordedDataPerActor.clear();
	recordedTimestampsPerActor.clear();

	const CapturyActor* actors;
	Captury_getActors(&actors);

	Captury_registerNewPoseCallback(newPose);

	if (!Captury_startStreaming(CAPTURY_STREAM_GLOBAL_POSES)) {
		mexPrintf("failed to start streaming\n");
		Captury_stopRecording();
		Captury_registerNewPoseCallback(nullptr);
		goto out;
	}

	recordingStartTime = Captury_startRecording();
	if (recordingStartTime == 0) {
		mexPrintf("failed to start recording\n");
		Captury_registerNewPoseCallback(nullptr);
		goto out;
	}

	ret = 1;
out:
	if (nlhs >= 1)
		plhs[0] = mxCreateLogicalScalar(ret);
}

static void stopRecording(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	Captury_stopRecording();
	Captury_stopStreaming();
	Captury_registerNewPoseCallback(nullptr);

	if (nlhs < 1) {
		mexPrintf("recorded data not stored\n");
		return;
	}

	if (recordedDataPerActor.size() == 1) {
		std::vector<float>& data = recordedDataPerActor.begin()->second;
		std::vector<uint64_t>& timestamps = recordedTimestampsPerActor.begin()->second;
		int numFrames = timestamps.size();

		int valuesPerFrame = data.size() / numFrames;

		int numTooEarly = 0;
		for (numTooEarly = 0; numTooEarly < numFrames && timestamps[numTooEarly] < recordingStartTime; ++numTooEarly)
			;
		numFrames -= numTooEarly;

		plhs[0] = mxCreateUninitNumericMatrix(valuesPerFrame, numFrames, mxSINGLE_CLASS, mxREAL);
		memcpy(mxGetData(plhs[0]), &data[numTooEarly * valuesPerFrame], numFrames * valuesPerFrame * sizeof(float));
		if (nlhs >= 2) {
			plhs[1] = mxCreateUninitNumericMatrix(1, numFrames, mxINT64_CLASS, mxREAL);
			std::vector<uint64_t>& timestamps = recordedTimestampsPerActor.begin()->second;
			memcpy(mxGetData(plhs[1]), &timestamps[numTooEarly], numFrames * sizeof(uint64_t));
		}
		if (nlhs >= 3) {
			const CapturyActor* actor = Captury_getActor(recordedTimestampsPerActor.begin()->first);
			std::vector<char*> names(actor->numJoints);
			for (int i = 0; i < actor->numJoints; ++i)
				names[i] = actor->joints[i].name;
			// const char** namesArray = &names[0];
			plhs[2] = mxCreateCharMatrixFromStrings(actor->numJoints, const_cast<const char**>(&names[0]));
		}
	} else {
		mwSize dims[3] = {0, 0, (mwSize)recordedTimestampsPerActor.size()};
		int numFrames = 0;
		int numTransforms = 0;
		if (nlhs >= 1) {
			for (const auto& d : recordedTimestampsPerActor) {
				int numTooEarly = 0;
				for (numTooEarly = 0; numTooEarly < numFrames && d.second[numTooEarly] < recordingStartTime; ++numTooEarly)
					;
				if (d.second.size() - numTooEarly > numFrames)
					numFrames = d.second.size() - numTooEarly;
			}
			for (const auto& d : numTransformsPerActor) {
				if (d.second > numTransforms)
					numTransforms = d.second;
			}

			dims[0] = numTransforms * 6;
			dims[1] = numFrames;
			plhs[0] = mxCreateNumericArray(3, dims, mxSINGLE_CLASS, mxREAL);
			float* reorganized = (float*)mxGetData(plhs[0]);
			int at = 0;
			for (const auto& d : recordedDataPerActor) {
				numTransforms = numTransformsPerActor[d.first];
				float* to = reorganized + at;
				std::vector<uint64_t>& timestamps = recordedTimestampsPerActor[d.first];
				int numTooEarly = 0;
				for (numTooEarly = 0; numTooEarly < timestamps.size() && timestamps[numTooEarly] < recordingStartTime; ++numTooEarly)
					;
				for (int i = numTooEarly; i < d.second.size(); ++i, to += dims[0])
					memcpy(to, d.second.data() + numTransforms * 6 * i, numTransforms * 6 * sizeof(float));
				at += dims[0] * dims[1];
			}
		}
		if (nlhs >= 2) {
			dims[0] = numFrames;
			dims[1] = (mwSize)recordedTimestampsPerActor.size();
			plhs[1] = mxCreateNumericArray(2, dims, mxUINT64_CLASS, mxREAL);
			uint64_t* data = (uint64_t*)mxGetData(plhs[1]);
			int at = 0;
			for (const auto& t : recordedTimestampsPerActor) {
				uint64_t* to = data + at;
				int numTooEarly = 0;
				for (numTooEarly = 0; numTooEarly < t.second.size() && t.second[numTooEarly] < recordingStartTime; ++numTooEarly)
					;
				memcpy(to, t.second.data() + numTooEarly, (t.second.size() - numTooEarly) * sizeof(uint64_t));
				at += dims[0];
			}
		}
		if (nlhs >= 3) {
			if (!recordedTimestampsPerActor.empty()) {
				const CapturyActor* actor = Captury_getActor(recordedTimestampsPerActor.begin()->first);
				if (actor != nullptr) {
					std::vector<char*> names(actor->numJoints);
					for (int i = 0; i < actor->numJoints; ++i)
						names[i] = actor->joints[i].name;
					plhs[2] = mxCreateCharMatrixFromStrings(actor->numJoints, const_cast<const char**>(&names[0]));
				} else {
					mwSize dims = 1;
					plhs[2] = mxCreateCharArray(1, &dims);
				}
			} else {
				mwSize dims = 1;
				plhs[2] = mxCreateCharArray(1, &dims);
			}
		}
	}
}

static void cleanup()
{
	Captury_disconnect();
}

static void selectShot(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])
{
	int ret = 0;
	if (nrhs != 2) {
		mexPrintf("the second argument needs to be a character array.\n");
		goto out;
	}

	if (!mxIsChar(prhs[1])) {
		mexPrintf("the second argument needs to be a character array.\n");
		goto out;
	}

	{
	std::vector<char> name(mxGetN(prhs[1]) + 1, '\0');
	strncpy(name.data(), mxArrayToString(prhs[1]), name.size() - 1);
	ret = Captury_setShotName(name.data());
	}
out:
	if (nlhs >= 1)
		plhs[0] = mxCreateLogicalScalar(ret);
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	mexAtExit(cleanup);

	if (nrhs < 1 || !mxIsScalar(prhs[0])) {
		mexPrintf("specify the type of action as the first parameter (0 - 7).\n0: [success] = connect(\"ip\"[, port])\n1: disconnect()\n2: [deltaTimeUs] = synchronizeTime()\n3: [success] = startRecording()\n4: [data] = stopRecording()\n5: [success] = selectShot(name)\n6: [posixTimeUs] = getTime()\n7: [deltaTimeUs] = getTimeOffset()\n");
		return;
	}

	int call = (int)mxGetScalar(prhs[0]);
	switch (call) {
	case 0:
		connect(nlhs, plhs, nrhs, prhs);
		break;
	case 1:
		Captury_disconnect();
		connected = false;
		break;
	case 2: // synchronize
		Captury_startTimeSynchronizationLoop();
		Captury_enablePrintf(0);
		if (nlhs >= 1)
			plhs[0] = mxCreateDoubleScalar(Captury_getTimeOffset());
		break;
	case 3:
		startRecording(nlhs, plhs, nrhs, prhs);
		break;
	case 4:
		stopRecording(nlhs, plhs, nrhs, prhs);
		break;
	case 5:
		selectShot(nlhs, plhs, nrhs, prhs);
		break;
	case 6:
		if (nlhs >= 1) {
			int64_t ts = Captury_getTime();
			mexPrintf("now %zd\n", ts);
			plhs[0] = mxCreateDoubleScalar(ts);
		}
		break;
	case 7:
		if (nlhs >= 1)
			plhs[0] = mxCreateDoubleScalar(Captury_getTimeOffset());
		break;
	case 8:
		Captury_enablePrintf(1);
		break;
	case 9:
		Captury_enablePrintf(0);
		break;
	}
}
